/*
 *  $Id: mxmpi_hrecv.c,v 1.13 2005/06/03 06:43:55 loic Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mxmpi.h"

#include "mpid.h"
#include "mpiddev.h"
#include "packets.h"
#include "reqalloc.h"

/***************************************************************************/
/*
 * Multi-protocol, Multi-device support for 2nd generation ADI.
 * This file has support for noncontiguous sends for systems that do not 
 * have native support for complex datatypes.
 */
/***************************************************************************/

void MPID_RecvDatatype( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int count, 
	struct MPIR_DATATYPE *dtype_ptr, 
	int src_lrank, 
	int tag, 
	int context_id, 
	MPI_Status *status, 
	int *error_code )
{
    MPI_Request  request;
    MPIR_RHANDLE rhandle;

    DEBUG_INIT_STRUCT(request,sizeof(rhandle));

    MPID_RecvInit( &rhandle );
    request =  (MPI_Request) &rhandle;

    /* rhandle.finish = 0; gets set in IrecvDatatype */
    *error_code = 0;
    MPID_IrecvDatatype( comm_ptr, buf, count, dtype_ptr, src_lrank, tag, 
			context_id, request, error_code );
    if (!*error_code) {
	MPID_RecvComplete( request, status, error_code );
    }
}

void MPID_IrecvDatatype( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int count, 
	struct MPIR_DATATYPE *dtype_ptr, 
	int src_lrank, 
	int tag, 
	int context_id, 
	MPI_Request request, 
	int *error_code )
{
  MPIR_RHANDLE    *rhandle;
  int             len;
  MPID_Msgrep_t   msgrep = MPID_MSGREP_RECEIVER;
  MPID_DO_HETERO(MPID_Msg_pack_t msgact = MPID_MSG_OK;)
  void            *mybuf;
  int             contig_size;
  mx_return_t rc;
  mx_segment_t seg[1];
  uint64_t match_val;
  uint64_t match_mask;
  MPID_DO_HETERO(int             src_grank);

  DEBUG_PRINT_ARGS("R starting IrecvDatatype");

  rhandle = &request->rhandle;

  /* Just in case; make sure that finish is 0 */
  rhandle->finish = 0;

  /* See if this is really contiguous */
  contig_size = MPIR_GET_DTYPE_SIZE(datatype,dtype_ptr);

  MPID_DO_HETERO(src_grank = (src_lrank >= 0) ? 
		 comm_ptr->lrank_to_grank[src_lrank] : src_lrank);
  MPID_DO_HETERO(MPID_Msg_rep( comm_ptr, src_grank, dtype_ptr, 
			       &msgrep, &msgact ));
  if (contig_size > 0
      MPID_DO_HETERO(&& msgact == MPID_MSG_OK)) {
      /* Just drop through into the contiguous send routine 
	 For packed data, the representation format is that in the
	 communicator.
       */
      len = contig_size * count;
      MPID_IrecvContig( comm_ptr, buf, len, src_lrank, tag, context_id, 
			request, error_code );
      return;
  }

  /* 
     Follow the same steps as IrecvContig, buf after creating a 
     temporary buffer to hold the incoming data in.
     */
  
  MPID_UnpackMessageSetup( count, dtype_ptr, comm_ptr, src_lrank, msgrep,
			   (void **)&mybuf, &len, error_code );
  if (*error_code) return;

  /* setup the request */
  /* Here we need to set up a different, special buffer if NOT 
     contiguous/homogeneous.  We'll also need a special complete 
     function to unpack the data
   */
  rhandle->len	 = len;
  rhandle->buf	 = mybuf;
  rhandle->start       = buf;
  rhandle->count       = count;
  rhandle->datatype    = dtype_ptr;
  MPIR_REF_INCR(dtype_ptr);
  rhandle->is_complete = 0;
  rhandle->wait        = 0;
  rhandle->test        = 0;
  rhandle->finish      = MPID_UnpackMessageComplete;

  /* setup segment list */
  seg[0].segment_ptr = mybuf;
  seg[0].segment_length = len;

/* Figure out match value and mask */
  if (tag == MPI_ANY_TAG) {
    if (src_lrank == MPI_ANY_SOURCE) {
      match_mask = MXMPI_MATCH_MASK_SRC_TAG;
      match_val = MXMPI_MATCH(context_id, 0, 0);
    } else {
      match_mask = MXMPI_MATCH_MASK_TAG;
      match_val = MXMPI_MATCH(context_id, src_lrank, 0);
    }
  } else {
    if (src_lrank == MPI_ANY_SOURCE) {
      match_mask = MXMPI_MATCH_MASK_SRC;
      match_val = MXMPI_MATCH(context_id, 0, tag);
    } else {
      match_mask = MX_MATCH_MASK_NONE;
      match_val = MXMPI_MATCH(context_id, src_lrank, tag);
    }
  }

  MXMPI_NOTE(("post recv 2 match=%x:%x mask=%x:%x\n",
		MX_U32(match_val), MX_L32(match_val),
	        MX_U32(match_mask), MX_L32(match_mask)));

  /* post the receive */
#if MX_INTERP
  if (interp_file != NULL) {
    fprintf(interp_file, "recv r%x e1 %d 0x%x 0x%x %s 1\n",
      (unsigned int)&rhandle->mx_req_handle,
      len,
      MX_U32(match_val), MX_L32(match_val),
      (match_mask == MX_MATCH_MASK_NONE) ? "none" :
	(match_mask == MXMPI_MATCH_MASK_SRC) ? "b" :
	(match_mask == MXMPI_MATCH_MASK_TAG) ? "c" : "bc");
    fflush(interp_file);
  }
#endif
  rc = mx_irecv(mxmpi.my_endpoint,
      seg, 1,
      match_val, match_mask,
      request,
      &rhandle->mx_req_handle);

  /* print something about the result */
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "[%d]: mx_irecv 2 failed: %s\n", 
	MPID_MyWorldRank,
	mx_strerror(rc));
  }

  rhandle->is_complete = 0;

  MXMPI_NOTE(("recv 2 posted handle=%x, req=%x\n",
       request->rhandle.mx_req_handle,
       (unsigned int)request));
}
