/*
 *
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mxmpi.h"

#include "mpid.h"
#include "mxmpi_inline.h"

/***************************************************************************/
/*
 * Multi-protocol, Multi-device support for 2nd generation ADI.
 * We start with support for blocking, contiguous sends.
 * Note the 'msgrep' field; this gives a hook for heterogeneous systems
 * which can be ignored on homogeneous systems.
 */
/***************************************************************************/
void MPID_SendContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int len, 
	int src_lrank, 
	int tag, 
	int context_id, 
	int dest_grank, 
	MPID_Msgrep_t msgrep, 
	int *error_code )
{
  mx_return_t rc;
  mx_segment_t seg;
  uint64_t match_send;
  mx_request_t mx_send_handle;
  mx_status_t stat;
  uint32_t result;

  /* Make sure length and pointer are good */
  if (buf == 0 && len > 0) {
      *error_code = MPI_ERR_BUFFER;
      return;
  }

  /* Build segment list */
  seg.segment_ptr = buf;
  seg.segment_length = len;

  /* Build the matching info */
  match_send = MXMPI_MATCH(context_id, src_lrank, tag);

  MXMPI_NOTE(("post send, d=%d, match=%x:%x, len = %d\n",
	  dest_grank,
	  MX_U32(match_send),
	  MX_L32(match_send),
	  len));

  /* Post the send */
#if MX_INTERP
if (interp_file) {
  fprintf(interp_file, "send s%x e1 t%d %d 0x%x 0x%x 1\n", 
	    (unsigned int)&mx_send_handle,
	    dest_grank,
	    len,
	    MX_U32(match_send),
	    MX_L32(match_send));
  fflush(interp_file);
}
#endif

  rc = mx_isend(mxmpi.my_endpoint,
      &seg, 1,
      mxmpi.addrs[dest_grank].addr,
      match_send,
      NULL,
      &mx_send_handle);
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "[%d] error from mx_isend: %s\n", 
	MPID_MyWorldRank, mx_strerror(rc));
    mxmpi_abort(0);
  }

  MXMPI_NOTE(("send posted, waiting, d=%d, len = %d, h=%x\n", 
      dest_grank, len, mx_send_handle));

  /* await completion */
#if MX_INTERP
if (interp_file != NULL) {
  fprintf(interp_file, "test s%x\n",
	    (unsigned int)&mx_send_handle);
  fflush(interp_file);
}
#endif

  while (1) {
    rc = mxmpi.test_or_wait(mxmpi.my_endpoint, &mx_send_handle, &stat, &result);

    if (rc != MX_SUCCESS) {
      fprintf(stderr, "[%d] error from test_or_wait for send 1: %s\n", 
	  MPID_MyWorldRank, mx_strerror(rc));
      mxmpi_abort(0);
    }

    if (result) break;
  }

  /* propagate any error code */
  *error_code = mxmpi_status_to_mpi_code(stat.code);

  MXMPI_NOTE(("send complete, h=%x\n", mx_send_handle));

#if 0
  *error_code = (*(fcn))( buf, len, src_lrank, tag, context_id, dest_grank, 
			  msgrep );
#endif
}

void MPID_IsendContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int len, 
	int src_lrank, 
	int tag, 
	int context_id, 
	int dest_grank, 
	MPID_Msgrep_t msgrep, 
	MPI_Request request, 
	int *error_code )
{
    mx_return_t rc;
    mx_segment_t seg;
    uint64_t match_send;

    /* Make sure length and pointer are good */
    if (buf == 0 && len > 0) {
	*error_code = MPI_ERR_BUFFER;
	return;
    }

    /* Build segment list */
    seg.segment_ptr = buf;
    seg.segment_length = len;

    /* Build the matching info */
    match_send = MXMPI_MATCH(context_id, src_lrank, tag);

    MXMPI_NOTE(("posting send, d=%d, match=%x:%x, len=%d\n",
                  dest_grank,
		  MX_U32(match_send), MX_L32(match_send),
		  len));

    /* Post the send */
#if MX_INTERP
  if (interp_file != NULL) {
    fprintf(interp_file, "send s%x e1 t%d %d 0x%x 0x%x 1\n", 
	      (unsigned int)&request->shandle.mx_req_handle,
	      dest_grank,
	      len,
	      MX_U32(match_send),
	      MX_L32(match_send));
    fflush(interp_file);
  }
#endif
    rc = mx_isend(mxmpi.my_endpoint,
	&seg, 1,
	mxmpi.addrs[dest_grank].addr,
	match_send,
	request,
	&request->shandle.mx_req_handle);
    if (rc != MX_SUCCESS) {
      fprintf(stderr, "[%d] error from mx_isend: %s\n", 
	  MPID_MyWorldRank, mx_strerror(rc));
      mxmpi_abort(0);
    }

    MXMPI_NOTE(("send posted, d=%d, len = %d, match=%x:%x, h=%x, sh=%x\n",
	dest_grank, len,
	MX_U32(match_send), MX_L32(match_send),
	request->shandle.mx_req_handle,
	&request->shandle));

    request->shandle.is_complete = 0;

#if 0
    MPID_Device *dev = MPID_devset->dev[dest_grank];
    int (*fcn) ( void *, int, int, int, int, int, MPID_Msgrep_t, 
		 MPIR_SHANDLE * );

    /* The one error test that makes sense here */
    if (buf == 0 && len > 0) {
	*error_code = MPI_ERR_BUFFER;
	return;
    }

    /* Just in case; make sure that finish is 0 */
    request->shandle.finish = 0;

    /* Choose the function based on the message length in bytes */
    if (len < dev->long_len)
	fcn = dev->short_msg->isend;
    else if (len < dev->vlong_len && MPID_FLOW_MEM_OK(len,dest_grank)) 
	fcn = dev->long_msg->isend;
    else
	fcn = dev->vlong_msg->isend;
    DEBUG_TEST_FCN(fcn,"dev->proto->isend");
    *error_code = (*(fcn))( buf, len, src_lrank, tag, context_id, dest_grank, 
			    msgrep, (MPIR_SHANDLE *)request );
#endif
}


/* Bsend is just a test for short send */
void MPID_BsendContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int len, 
	int src_lrank, 
	int tag, 
	int context_id, 
	int dest_grank, 
	MPID_Msgrep_t msgrep, 
	int *error_code )
{
MXMPI_NOTIMPL();
#if 0
    MPID_Device *dev = MPID_devset->dev[dest_grank];
    int rc;

    if (len < dev->long_len) {
	DEBUG_TEST_FCN(dev->short_msg->send,"dev->short->send");
	rc = (*dev->short_msg->send)( buf, len, src_lrank, tag, context_id, 
				      dest_grank, msgrep );
    }
    else
	rc = MPIR_ERR_MAY_BLOCK;
    *error_code = rc;
#endif
}

int MPID_SendIcomplete( 
	MPI_Request request,
	int         *error_code)
{
  mx_status_t stat;
  mx_return_t rc;
  uint32_t result;

  MXMPI_NOTE(("1 checking send complete h=%d\n",
    request->shandle.mx_req_handle));

  /* If request marked complete, just believe it */
  if (request->shandle.is_complete) {
    if (request->shandle.mx_req_handle != 0) {
      fprintf(stderr, "[%d] is_complete, but mx_req not NULL!\n",
      	MPID_MyWorldRank);
      abort();
    }

  } else {

    /* check for completion */
    rc = mx_test(mxmpi.my_endpoint,
		 &request->shandle.mx_req_handle, &stat, &result);

    if (rc != MX_SUCCESS) {
      fprintf(stderr, "[%d] test_or_wait error, stat=%s\n", 
	  MPID_MyWorldRank, mx_strerror(rc));
      mxmpi_abort(0);
    }

    if (result) {

#if MX_INTERP
      if (interp_file != NULL) {
	fprintf(interp_file, "test s%x\n",
		  (unsigned int)&request->shandle.mx_req_handle);
	fflush(interp_file);
      }
#endif

      request->shandle.is_complete = 1;
      request->shandle.mx_req_handle = 0;
      request->shandle.s.MPI_ERROR = mxmpi_status_to_mpi_code(stat.code);

      /* perform any finishing work needed */
      if (request->shandle.finish) {
	  (request->shandle.finish)( &request->shandle );
      }
    }
  }

  /* If request complete, finish it off */
  if (request->shandle.is_complete) {
    MXMPI_NOTE(("send complete, h=%x, sh=%x\n",
      request->shandle.mx_req_handle, &request->shandle));

    *error_code = request->shandle.s.MPI_ERROR;

  } else {
    MXMPI_NOTE(("send INCOMPLETE, h=%x\n", request->shandle.mx_req_handle));
  }

  return (request->shandle.is_complete);

#if 0
    MPIR_SHANDLE *shandle = &request->shandle;
    int lerr;

    if (shandle->is_complete) {
	if (shandle->finish) 
	    (shandle->finish)( shandle );
	return 1;
    }
    if (shandle->test) 
	*error_code = 
	    (*shandle->test)( shandle );
    else {
	/* The most common case is a check device loop */
	MPID_Device *dev;
	dev = MPID_devset->dev_list;
	while (dev) {
	    lerr = (*dev->check_device)( dev, MPID_NOTBLOCKING );
	    if (lerr > 0) {
		*error_code = lerr;
		return 0;
	    }
	    dev = dev->next;
	}
    }
    if (shandle->is_complete && shandle->finish) 
	(shandle->finish)( shandle );
    return shandle->is_complete;
#endif
}

void MPID_SendComplete( 
	MPI_Request request,
	int         *error_code)
{
  mx_status_t stat;
  mx_return_t rc;
  uint32_t result;

  MXMPI_NOTE(("awaiting send 3 complete h=%x, sh=%x\n",
    request->shandle.mx_req_handle, &request->shandle));

  /* If request marked complete, just believe it */
  if (request->shandle.is_complete) {
    if (request->shandle.mx_req_handle != 0) {
      fprintf(stderr, "[%d] is_complete, but send mx_req not NULL!\n",
	MPID_MyWorldRank);
      abort();
    }

  } else {

    /* Loop until done */
    do {


      /* check for completion */
      rc = mxmpi.test_or_wait(mxmpi.my_endpoint, &request->shandle.mx_req_handle,
			      &stat, &result);

      if (rc != MX_SUCCESS) {
	fprintf(stderr, "[%d] error from test_or_wait for send 3: %s\n", 
	    MPID_MyWorldRank, mx_strerror(rc));
	mxmpi_abort(0);
      }

      if (result) {
	request->shandle.s.MPI_ERROR = mxmpi_status_to_mpi_code(stat.code);
	request->shandle.is_complete = 1;
	request->shandle.mx_req_handle = 0;
      }
    } while (!request->shandle.is_complete);

#if MX_INTERP
    if (interp_file != NULL) {
      fprintf(interp_file, "test s%x\n",
		(unsigned int)&request->shandle.mx_req_handle);
      fflush(interp_file);
    }
#endif

    /* perform any finishing work needed */
    if (request->shandle.finish) {
	(request->shandle.finish)( &request->shandle );
    }
  }

  MXMPI_NOTE(("send complete, h=%x, m=%x:%x sh=%x\n",
	request->shandle.mx_req_handle,
	stat.unmaskable,
	stat.maskable,
	&request->shandle));

  *error_code = request->shandle.s.MPI_ERROR;

#if 0
    MPIR_SHANDLE *shandle = &request->shandle;
    int          lerr;

    /* The 'while' is at the top in case the 'wait' routine is changed
       by one of the steps.  This happens, for example, in the Rendezvous
       Protocol */
    DEBUG_PRINT_MSG( "Entering while !shandle->is_complete" );
    while (!shandle->is_complete) {
	if (shandle->wait) 
	    *error_code = 
		(*shandle->wait)( shandle );
	else {
	    /* The most common case is a check device loop until it is
	       complete. */
	    MPID_Device *dev;

	    if (MPID_devset->ndev_list == 1) {
		dev = MPID_devset->dev_list;
		if (!shandle->is_complete) {
		    lerr = (*dev->check_device)( dev, MPID_BLOCKING );
		    if (lerr > 0) {
			*error_code = lerr;
			return;
		    }
		}
	    }
	    else {
		if (!shandle->is_complete) {
		    dev = MPID_devset->dev_list;
		    while (dev) {
			lerr = (*dev->check_device)( dev, MPID_NOTBLOCKING );
			if (lerr > 0) {
			    *error_code = lerr;
			    return;
			}
			dev = dev->next;
		    }
		}
	    }
	}
    }
    DEBUG_PRINT_MSG( "Leaving while !shandle->is_complete" );
    if (shandle->finish) 
	(shandle->finish)( shandle );
#endif
}
