/*
 *  $Id: mxmpi_ssend.c,v 1.12 2007/01/31 19:02:42 loic Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mxmpi.h"
#include "mpid.h"
#include "mpiddev.h"
#include "mxmpi_inline.h"


/***************************************************************************/
/*
 * Multi-protocol, Multi-device support for 2nd generation ADI.
 * We start with support for blocking, contiguous sends.
 * Note the 'msgrep' field; this gives a hook for heterogeneous systems
 * which can be ignored on homogeneous systems.
 * 
 * For the synchronous send, we always use a Rendezvous send.
 */
/***************************************************************************/

void MPID_SsendContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int len, 
	int src_lrank, 
	int tag, 
	int context_id, 
	int dest_grank, 
	MPID_Msgrep_t msgrep, 
	int *error_code )
{
  mx_return_t rc;
  mx_segment_t seg;
  uint64_t match_send;
  mx_request_t mx_send_handle;
  uint32_t result;
  mx_status_t stat;

  /* Make sure length and pointer are good */
  if (buf == 0 && len > 0) {
      *error_code = MPI_ERR_BUFFER;
      return;
  }

  /* Build segment list */
  seg.segment_ptr = buf;
  seg.segment_length = len;

  /* Build the matching info */
  match_send = MXMPI_MATCH(context_id, src_lrank, tag);

  MXMPI_NOTE(("[%d] post ssend, d=%d, match=%x:%x, len = %d\n",
	  dest_grank,
	  MX_U32(match_send),
	  MX_L32(match_send),
	  len));

  /* Post the send */
#if MX_INTERP
  if (interp_file) {
    fprintf(interp_file, "ssend s%x e1 t%d %d 0x%x 0x%x 1\n", 
	      (unsigned int)&mx_send_handle,
	      dest_grank,
	      len,
	      MX_U32(match_send),
	      MX_L32(match_send));
    fflush(interp_file);
  }
#endif
  rc = mx_issend(mxmpi.my_endpoint,
      &seg, 1,
      mxmpi.addrs[dest_grank].addr,
      match_send,
      NULL,
      &mx_send_handle);
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "[%d]: error from mx_issend: %s\n", 
	MPID_MyWorldRank, mx_strerror(rc));
    mxmpi_abort(0);
  }

  MXMPI_NOTE(("[%d] ssend posted, waiting, d=%d, len = %d, h=%x\n",
      dest_grank, len, mx_send_handle));

  /* await completion */
#if MX_INTERP
if (interp_file != NULL) {
  fprintf(interp_file, "test s%x\n",
	    (unsigned int)&mx_send_handle);
  fflush(interp_file);
}
#endif
  while (1) {
    rc = mxmpi.test_or_wait(mxmpi.my_endpoint, &mx_send_handle, &stat, &result);

    if (rc != MX_SUCCESS) {
      fprintf(stderr, "[%d]: error from test_or_wait for send 1: %s\n", 
	  MPID_MyWorldRank, mx_strerror(rc));
      mxmpi_abort(0);
    }

    if (result) break;
  }

  *error_code = mxmpi_status_to_mpi_code(stat.code);

  MXMPI_NOTE(("ssend complete, h=%x\n", mx_send_handle));
}

void MPID_IssendContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int len, 
	int src_lrank, 
	int tag, 
	int context_id, 
	int dest_grank, 
	MPID_Msgrep_t msgrep, 
	MPI_Request request, 
	int *error_code )
{
    mx_return_t rc;
    mx_segment_t seg;
    uint64_t match_send;

    /* Make sure length and pointer are good */
    if (buf == 0 && len > 0) {
	*error_code = MPI_ERR_BUFFER;
	return;
    }

    /* Build segment list */
    seg.segment_ptr = buf;
    seg.segment_length = len;

    /* Build the matching info */
    match_send = MXMPI_MATCH(context_id, src_lrank, tag);

    MXMPI_NOTE(("posting issend, d=%d, match=%x:%x, len=%d\n",
                  dest_grank,
		  MX_U32(match_send), MX_L32(match_send),
		  len));

    /* Post the send */
#if MX_INTERP
  if (interp_file != NULL) {
    fprintf(interp_file, "ssend s%x e1 t%d %d 0x%x 0x%x 1\n", 
	      (unsigned int)&request->shandle.mx_req_handle,
	      dest_grank,
	      len,
	      MX_U32(match_send),
	      MX_L32(match_send));
    fflush(interp_file);
  }
#endif
    rc = mx_issend(mxmpi.my_endpoint,
	&seg, 1,
	mxmpi.addrs[dest_grank].addr,
	match_send,
	request,
	&request->shandle.mx_req_handle);
    if (rc != MX_SUCCESS) {
      fprintf(stderr, "[%d] error from mx_issend: %s\n", 
	  MPID_MyWorldRank, mx_strerror(rc));
      mxmpi_abort(0);
    }

    MXMPI_NOTE(("issend posted, d=%d, len = %d, match=%x:%x, h=%x, sh=%x\n",
	dest_grank, len,
	MX_U32(match_send), MX_L32(match_send),
	request->shandle.mx_req_handle,
	&request->shandle));

    request->shandle.is_complete = 0;
}
