/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#ifdef WIN32
#define strcasecmp stricmp
#endif

#define MX_STRONG_TYPES 0

#include "mxmpi.h"
#include "mxmpi_smpi.h"
#include "mpiddev.h"
#include "mpimem.h"
#include "bnr.h"


static void 
mxmpi_getenv (const char *varenv, char **result, char *msg, 
	     unsigned int required)
{
  *result = (char *) getenv (varenv);
  if ((required == 1) && (*result == NULL))
    {
      fprintf (stderr, "<MPICH-MX> Error: Need to obtain %s in %s !\n",
	       msg, varenv);
      mxmpi_abort (0);
    }
}


static void
mxmpi_allocate_port (int *board_number)
{
  uint32_t board_id;
  uint32_t bid_len;
  uint32_t count;
  mx_status_t status;
  
  /* do we know which board to use ? */
  if (*board_number < 0)
    { 
      for (count = 0; count < 3; count++)
	{
	  status = mx_open_port(MX_BOARD_ANY, MXMPI_GLOBAL_MAGIC, MXMPI_LOCAL_MAGIC,
	                        0, &mxmpi_mx_port);
	  if (status == MX_SUCCESS)
	    {
	      bid_len = sizeof (board_id);
	      if (mx_get_attr(mxmpi_mx_port, "board_number",
		              &board_id, &bid_len) != MX_SUCCESS) {
		fprintf(stderr, "<MPICH-MX> Error: Cannot get board number.\n");
		mxmpi_abort(0);
	      }
	      *board_number = board_id;
	      return;
	    }
	}
    }
  else
    {
      /* open a port on the allocated board */
      board_id = *board_number;
      for (count = 0; count < 3; count++)
	{
	  status = mx_open_port(board_id, MXMPI_GLOBAL_MAGIC, MXMPI_LOCAL_MAGIC,
	                        0, &mxmpi_mx_port);
	  if (status == MX_SUCCESS)
	    {
	      return;
	    }
	}
    }

#if 0
  if (status == MX_INCOMPATIBLE_LIB_AND_DRIVER)
    {
      fprintf (stderr, "[%d] Error: The MX driver and the lib used to "
	       "compile MPICH-MX are not compatible ! Please recompile "
	       "MPICH-MX with a lib compatible with the current MX "
	       "driver.\n", MPID_MyWorldRank);
      mxmpi_abort (0);
    }
#endif
  fprintf(stderr, "[%d] Error: mx_open_port() failed: %s\n",
          MPID_MyWorldRank, mx_strerror(status));
}


static void
mxmpi_allocate_world (unsigned int size)
{
  mxmpi.endpoints = (unsigned int *) calloc (size, sizeof (*mxmpi.endpoints));
  mxmpi.mpi_pids = (unsigned int *) calloc (size, sizeof (unsigned int));
  mxmpi.pending_sends = (unsigned int *) calloc (size, sizeof (unsigned int));
  mxmpi.dropped_sends = (unsigned int *) calloc (size, sizeof (unsigned int));
  mxmpi.host_names = (char **) calloc (size, sizeof (char *));
  mxmpi.exec_names = (char **) calloc (size, sizeof (char *));
  
  mxmpi_malloc_assert (mxmpi.endpoints, "mxmpi_getconf", "malloc: endpoints");
  mxmpi_malloc_assert (mxmpi.mpi_pids, "mxmpi_getconf", "malloc: mpi_pids");
  mxmpi_malloc_assert (mxmpi.pending_sends, "mxmpi_getconf", 
		      "malloc: pending_sends");
  mxmpi_malloc_assert (mxmpi.dropped_sends, "mxmpi_getconf", 
		      "malloc: dropped_sends");
  mxmpi_malloc_assert (mxmpi.host_names, "mxmpi_getconf", "malloc: host_names");
  mxmpi_malloc_assert (mxmpi.exec_names, "mxmpi_getconf", "malloc: exec_names");
}


#define MXMPI_SOCKET_BUFFER_SIZE 128*1024

static void
mxmpi_getconf (void)
{
  char *mxmpi_eager, *mxmpi_shmem, *mxmpi_recvmode;
  unsigned int i, j;
  int board_id, k;
  uint32_t alen;
  int8_t ep_buf[(MX_SIZEOF_ENDPOINT + 1) * 2];
  
  setbuf (stdout, NULL);
  setbuf (stderr, NULL);

  if (getenv ("MAN_MPD_FD") != NULL)
    {
      char attr_buffer[BNR_MAXATTRLEN];
      char val_buffer[BNR_MAXVALLEN];
      /* MPD to spawn processes */
      char my_hostname[256];
      char *hostnames;
      BNR_Group bnr_group;
      
      /* MPD to spawn processes */
      mxmpi.mpd = 1;
      i = BNR_Init ();
      i = BNR_Get_group (&bnr_group);
      i = BNR_Get_rank (bnr_group, &MPID_MyWorldRank);
      i = BNR_Get_size (bnr_group, &MPID_MyWorldSize);
      
      /* data allocation */
      mxmpi_allocate_world (MPID_MyWorldSize);
      hostnames = (char *) malloc (MPID_MyWorldSize * 256 * sizeof (char *));
      mxmpi_malloc_assert (hostnames, "mxmpi_getconf", "malloc: hostnames");

      /* open a port */
      board_id = -1;

      mxmpi_allocate_port (&board_id);
      if (mxmpi_mx_port == NULL)
	{
	  fprintf (stderr, "[%d] Error: Unable to open a MX port !\n", 
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
      
      /* get my MX endpoint */
      alen = sizeof (mxmpi.my_endpoint);
      if (mx_get_attr(mxmpi_mx_port,
	              "my_endpoint",
		      &(mxmpi.my_node_id),
		      &alen) != MX_SUCCESS)
	{
	  fprintf (stderr, "[%d] Error: Unable to get MX local endpoint !\n", 
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
  
      /* build the data to send to master */
      bzero (val_buffer, BNR_MAXVALLEN * sizeof (char));
      gethostname(my_hostname, sizeof (my_hostname));
      sprintf (val_buffer, "< %s:%s >\n", 
	       mx_endpoint_to_text(mxmpi.my_endpoint), my_hostname);
      
      /* put our information */
      bzero (attr_buffer, BNR_MAXATTRLEN * sizeof (char));
      sprintf (attr_buffer, "MPICH-MX data [%d]\n", MPID_MyWorldRank);
      i = BNR_Put (bnr_group, attr_buffer, val_buffer, -1);
      
      /* get other processes data */
      i = BNR_Fence (bnr_group);
      for (j = 0; j < MPID_MyWorldSize; j++)
	{
	  bzero (attr_buffer, BNR_MAXATTRLEN * sizeof (char));
	  sprintf (attr_buffer, "MPICH-MX data [%d]\n", j);
	  i = BNR_Get (bnr_group, attr_buffer, val_buffer);
	  
	  /* decrypt data */
	  if (sscanf (val_buffer, "< %s:%s >",
		      ep_buf,
		      &(hostnames[j*256])) != 2)
	    {
	      fprintf (stderr, "[%d] Error: unable to decode data "
		       "from %d !\n", MPID_MyWorldRank, j);
	      mxmpi_abort (0);
	    }
	  mxmpi.endpoints[j] = mx_text_to_endpoint(ep_buf);
	}
      
      /* compute the local mapping */
      smpi.num_local_nodes = 0;
      for (j = 0; j < MPID_MyWorldSize; j++)
	{
	  if (strcmp (my_hostname, &(hostnames[j*256])) == 0)
	    {
	      if (j == MPID_MyWorldRank)
		{
		  smpi.my_local_id = smpi.num_local_nodes;
		}
	      smpi.local_nodes[j] = smpi.num_local_nodes;
	      smpi.num_local_nodes++;
	    }
	  else
	    {
	      smpi.local_nodes[j] = -1;
	    }
	}
      free (hostnames);
    }
  else
    {
      char *mxmpi_magic, *mxmpi_master, *mxmpi_port1, *mxmpi_port2;
      char *mxmpi_id, *mxmpi_np, *mxmpi_board;
      char buffer[MXMPI_SOCKET_BUFFER_SIZE];
      char temp[64];
      unsigned int count, magic_number, master_port1, master_port2;
      int mxmpi_sockfd;
      struct hostent *master;
      
      /* mpirun with sockets */
      mxmpi.mpd = 0;
      gethostname(temp, sizeof (temp)-1);
      mxmpi_getenv ("MXMPI_MAGIC", &mxmpi_magic, "the job magic number", 1);
      mxmpi_getenv ("MXMPI_MASTER", &mxmpi_master, "the master's hostname", 1);
      mxmpi_getenv ("MXMPI_PORT1", &mxmpi_port1, "the master's port 1 number", 1);
      mxmpi_getenv ("MXMPI_PORT2", &mxmpi_port2, "the master's port 2 number", 1);
      mxmpi_getenv ("MXMPI_ID", &mxmpi_id, "the MPI ID of the process", 1);
      mxmpi_getenv ("MXMPI_NP", &mxmpi_np, "the number of MPI processes", 1);
      mxmpi_getenv ("MXMPI_BOARD", &mxmpi_board, "the specified board", 1);
      
      if (sscanf (mxmpi_magic, "%d", &magic_number) != 1)
	{
	  fprintf (stderr, "<MPICH-MX> Error on %s: Bad magic number "
		   "(MXMPI_MAGIC is %s) !\n", temp, mxmpi_magic);
	  mxmpi_abort (0);
	}
      mxmpi.magic = magic_number;
      
      if ((sscanf (mxmpi_np, "%d", &MPID_MyWorldSize) != 1)
	  || (MPID_MyWorldSize < 0))
	{
	  fprintf (stderr, "<MPICH-MX> Error on %s: Bad number of processes "
		   "(MXMPI_NP is %s) !\n", temp, mxmpi_np);
	  mxmpi_abort (0);
	}
      
      if ((sscanf (mxmpi_id, "%d", &MPID_MyWorldRank) != 1)
	  || (MPID_MyWorldRank < 0) || (MPID_MyWorldRank >= MPID_MyWorldSize))
	{
	  fprintf (stderr, "<MPICH-MX> Error on %s: Bad MPI ID "
		   "(MXMPI_ID is %s, total number of MPI processes is %d) !\n", 
		   temp, mxmpi_np, MPID_MyWorldSize);
	  mxmpi_abort (0);
	}
  
      if (sscanf (mxmpi_port1, "%d", &master_port1) != 1)
	{
	  fprintf (stderr, "<MPICH-MX> Error on %s: Bad master port 1 number "
		   "(MXMPI_PORT1 is %s) !\n", temp, mxmpi_port1);
	  mxmpi_abort (0);
	}

      if (sscanf (mxmpi_port2, "%d", &master_port2) != 1)
	{
	  fprintf (stderr, "<MPICH-MX> Error on %s: Bad master port 2 number "
		   "(MXMPI_PORT2 is %s) !\n", temp, mxmpi_port2);
	  mxmpi_abort (0);
	}

      if (sscanf (mxmpi_board, "%d", &board_id) != 1)
	{
	  fprintf (stderr, "<MPICH-MX> Error on %s: Bad board ID "
		   "(MXMPI_BOARD is %s) !\n", temp, mxmpi_board);
	  mxmpi_abort (0);
	}

      /* data allocation */
      mxmpi_allocate_world (MPID_MyWorldSize);

      /* open a port */
      port_id = 0;
      mxmpi_allocate_port (&board_id);
      if (mxmpi_mx_port == NULL)
	{
	  fprintf (stderr, "[%d] Error: Unable to open a MX port !\n", 
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}

      /* get my MX endpoint */
      alen = sizeof (mxmpi.my_endpoint);
      if (mx_get_attr(mxmpi_mx_port,
	              "my_endpoint",
		      &(mxmpi.my_node_id),
		      &alen) != MX_SUCCESS)
	{
	  fprintf (stderr, "[%d] Error: Unable to get MX local endpoint !\n", 
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
  
      /* get a socket */
      mxmpi_sockfd = socket (AF_INET, SOCK_STREAM, 0);
      if (mxmpi_sockfd < 0)
	{
	  fprintf (stderr, "[%d] Error: Unable to open a socket !\n", 
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
  
      /* get the master's IP address */
      master = gethostbyname (mxmpi_master);
      if (master == NULL)
	{ 
	  fprintf (stderr, "[%d] Error: Unable to translate "
		   "the hostname of the master (%s)!\n", 
		   MPID_MyWorldRank, mxmpi_master);
	  mxmpi_abort (0);
	}
  
      /* connect to the master */
      bzero ((char *) (&(mxmpi.master_addr)), sizeof (mxmpi.master_addr));
      mxmpi.master_addr.sin_family = AF_INET;
      mx_bcopy ((char *) (master->h_addr), 
		(char *) (&(mxmpi.master_addr.sin_addr)),
		master->h_length);
      mxmpi.master_addr.sin_port = htons (master_port1);
      i = MXMPI_INIT_TIMEOUT;
      while (connect (mxmpi_sockfd, (struct sockaddr *) (&(mxmpi.master_addr)), 
		      sizeof (mxmpi.master_addr)) < 0)
	{
	  usleep (1000);
	  if (i == 0)
	    {
	      fprintf (stderr, "[%d] Error: Unable to connect to "
		       "the master !\n", MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
	  i--;
	}
  
      /* send the magic:ID:endpoint:pid used to the master */
      count = 0;
      sprintf (buffer, "<<<%d:%d:%s:%d>>>\n", magic_number, 
	       MPID_MyWorldRank, mx_endpoint_to_text(mxmpi.my_endpoint),
	       (int) getpid ());
      while (count < strlen (buffer))
	{
	  k = write (mxmpi_sockfd, &(buffer[count]), strlen (buffer) - count);
	  if (k < 0)
	    {
	      fprintf (stderr, "[%d] Error: write to socket failed !\n", 
		       MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
	  count += k;
	}
      close (mxmpi_sockfd);
      
      /* get another socket */
      mxmpi_sockfd = socket (AF_INET, SOCK_STREAM, 0);
      if (mxmpi_sockfd < 0)
	{
	  fprintf (stderr, "[%d] Error: Unable to open a socket (2)!\n", 
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
      
      /* re-connect to the master */
      mxmpi.master_addr.sin_port = htons (master_port2);
      i = MXMPI_INIT_TIMEOUT;
      while (connect (mxmpi_sockfd, (struct sockaddr *) (&(mxmpi.master_addr)), 
		      sizeof (mxmpi.master_addr)) < 0)
	{
	  usleep (1000);
	  if (i == 0)
	    {
	      fprintf (stderr, "[%d] Error: Unable to connect(2) to "
		       "the master !\n", MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
	  i--;
	}
      
      /* send the a small authentification magic:ID to the master */
      count = 0;
      sprintf (buffer, "<->%d:%d<->\n", magic_number, MPID_MyWorldRank);
      while (count < strlen (buffer))
	{
	  k = write (mxmpi_sockfd, &(buffer[count]), strlen (buffer) - count);
	  if (k < 0)
	    {
	      fprintf (stderr, "[%d] Error: write to socket (2) failed !\n", 
		       MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
	  count += k;
	}

      /* Get the whole MX mapping from the master */
      count = 0;
      bzero (buffer, MXMPI_SOCKET_BUFFER_SIZE * sizeof(char));
      while (strstr (buffer, "]]]") == NULL)
	{
	  k = read (mxmpi_sockfd, &(buffer[count]), 
		    MXMPI_SOCKET_BUFFER_SIZE - count);
	  if (k < 0)
	    {
	      fprintf (stderr, "[%d] Error: read from socket failed !\n", 
		       MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
	  count += k;
	}
      close (mxmpi_sockfd);
  
      /* check the initial marker */
      j = 0;
      if (strncmp (&(buffer[j]), "[[[", 3) != 0)
	{
	  fprintf (stderr, "[%d] Error: bad format on data from master !\n",
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}

      /* Decrypt the global mapping */
      j += 3;
      for (i = 0; i < MPID_MyWorldSize; i++)
	{
	  if (sscanf (&(buffer[j]), "<%s>", &ep_buf) != 1)
	    {
	      fprintf (stderr, "[%d] Error: unable to decode data "
		       "from master !\n", MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
	  mxmpi.endpoints[j] = mx_text_to_endpoint(ep_buf);
      
	  sprintf (temp, "<%s>", mx_endpoint_to_text(mxmpi.endpoints[j]));
	  j += strlen (temp);

	  smpi.local_nodes[i] = -1;
	}

      /* check the marker between global map and local map */  
      if (strncmp (&(buffer[j]), "|||", 3) != 0)
	{
	  fprintf (stderr, "[%d] Error: bad format on data from master !\n",
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}

      /* decrypt the local mapping */
      j += 3;
      smpi.num_local_nodes = 0;
      while (strncmp (&(buffer[j]), "]]]", 3) != 0)
	{ 
	  if (sscanf (&(buffer[j]), "<%d>", &i) != 1)
	    {
	      fprintf (stderr, "[%d] Error: unable to decode master data !\n",
		       MPID_MyWorldRank);
	      mxmpi_abort (0);
	    }
      
	  if (i == MPID_MyWorldRank)
	    {
	      smpi.my_local_id = smpi.num_local_nodes;
	    }
	  smpi.local_nodes[i] = smpi.num_local_nodes;
	  smpi.num_local_nodes++;
      
	  sprintf (temp, "<%d>", i);
	  j += strlen (temp);
	}

      /* check size of the data from the master */
      j += 3;
      if (j != count)
	{
	  fprintf (stderr, "[%d] Error: amount of data from master !\n",
		   MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
    }

  /* check consistency */
  if ((mxmpi.port_ids[MPID_MyWorldRank] != port_id) 
      || (mxmpi.board_ids[MPID_MyWorldRank] != board_id)
      || (mxmpi.node_ids[MPID_MyWorldRank] != mxmpi.my_node_id))
    {
      fprintf (stderr, "[%d] Error: inconsistency in collected data !\n", 
	       MPID_MyWorldRank);
      mxmpi_abort (0);
    }

  mxmpi_getenv ("MXMPI_EAGER", &mxmpi_eager, NULL, 0);
  mxmpi_getenv ("MXMPI_SHMEM", &mxmpi_shmem, NULL, 0);
  mxmpi_getenv ("MXMPI_RECV", &mxmpi_recvmode, NULL, 0);

  /* Set the EAGER/Rendez-vous protocols threshold */
  if (mxmpi_eager == NULL)
    {
      mxmpi.eager_size = MXMPI_EAGER_SIZE_DEFAULT;
    }
  else
    {
      mxmpi.eager_size = strtoul(mxmpi_eager, NULL, 10);
      if (mxmpi.eager_size < 128)
	{
	  mxmpi.eager_size = 128;
	}
      
      if (mxmpi.eager_size > MXMPI_MAX_PUT_LENGTH)
	{
	  mxmpi.eager_size = MXMPI_MAX_PUT_LENGTH;
	}
    }

  /* set the MX receive mode */
  /* reese XXX - make mx_receive_mode a flag */
  if (mxmpi_recvmode == NULL)
    {
      mxmpi.mx_receive_mode = MXMPI_RXMODE_POLLING;
    }
  else
    {
      if (strcasecmp (mxmpi_recvmode, "polling") == 0)
	{
	  mxmpi.mx_receive_mode = MXMPI_RXMODE_POLLING;
	}
      else
	{
	  if (strcasecmp (mxmpi_recvmode, "blocking") == 0)
	    {
	      mxmpi.mx_receive_mode = MXMPI_RXMODE_BLOCKING;
	    }
	  else
	    {
	      if (strcasecmp (mxmpi_recvmode, "hybrid") == 0)
		{
		  mxmpi.mx_receive_mode = MXMPI_RXMODE_BLOCKING;
		}
	      else
		{
		  mxmpi.mx_receive_mode = MXMPI_RXMODE_POLLING;
		}
	    }
	}
    }

  /* Set the shared memory support */
  if (mxmpi.mx_receive_mode == MXMPI_RXMODE_POLLING)
    {
      if (mxmpi_shmem == NULL)
	{
	  mxmpi.shmem = 1;
	}
      else
	{
	  if (strcmp (mxmpi_shmem, "1") == 0) 
	    {
	      mxmpi.shmem = 1;
	    }
	  else
	    {
	      mxmpi.shmem = 0;
	    }
	}
    }
  else
    {
      mxmpi.shmem = 0;
    }
}


/* This function fill the MPID_Config structure to describe 
   the multi-devices configuration */
MPID_Config *MPID_GetConfigInfo (int *argc, 
				 char ***argv, 
				 void *config, 
				 int  *error_code)
{
  MPID_Config * new_config = NULL;
  MPID_Config * return_config = NULL;
  int i, j;

  /* already configured ? */
  if (config != NULL)
      return (MPID_Config *)config;

  /* Get the MX mapping and the environnement variables */
  mxmpi_getconf();

  /* at least one local node : myself ! */ 
  mxmpi_debug_assert(smpi.num_local_nodes != 0);


  /* SELF DEVICE */
  /* if it's the first device, start the linked list of devices.
     Otherwise, add a new one at the end. */
  if (new_config == NULL)
    {
      new_config = (MPID_Config *)malloc(sizeof(MPID_Config));
      mxmpi_malloc_assert(new_config,
			 "MPID_GetConfigInfo",
			 "malloc: Self device config");
      return_config = new_config;
    }
  else
    {
      new_config->next = (MPID_Config *)malloc(sizeof(MPID_Config));
      mxmpi_malloc_assert(new_config->next,
			 "MPID_GetConfigInfo",
			 "malloc: Self device config");
      new_config = new_config->next;
    }
  /* we don't need to check if we need this device: we need this device ! */
  new_config->device_init = MPID_CH_InitSelfMsg;
  new_config->device_init_name = (char *)malloc(255*sizeof(char));
  mxmpi_malloc_assert(new_config->device_init_name,
		     "MPID_GetConfigInfo",
		     "malloc: Self device name");
  sprintf(new_config->device_init_name, "Self device");
  new_config->num_served = 1;
  new_config->granks_served = (int *)malloc(sizeof(int));
  mxmpi_malloc_assert(new_config->granks_served,
		     "MPID_GetConfigInfo",
		     "malloc: Self device map");
  new_config->granks_served[0] = MPID_MyWorldRank;
  new_config->next = NULL;
  
  
  /* SMP DEVICE */
  /* we don't need this device if there's only one process on this node */
  if (smpi.num_local_nodes > 1) 
    {
#if !MX_OS_VXWORKS
      if (mxmpi.shmem == 1)
	{
	  if (mxmpi.mx_receive_mode == mx_receive) 
	    {
	      if (new_config == NULL) 
		{
		  new_config = (MPID_Config *)malloc(sizeof(MPID_Config));
		  mxmpi_malloc_assert(new_config,
				     "MPID_GetConfigInfo",
				     "malloc: SMP device config");
		  return_config = new_config;
		}
	      else 
		{
		  new_config->next = (MPID_Config *)
		    malloc(sizeof(MPID_Config));
		  mxmpi_malloc_assert(new_config->next,
				     "MPID_GetConfigInfo",
				     "malloc: SMP device config");
		  new_config = new_config->next;
		}
	      new_config->device_init = MPID_SMP_InitMsgPass;
	      new_config->device_init_name = (char *)
		malloc(255 * sizeof(char));
	      mxmpi_malloc_assert(new_config->device_init_name,
				 "MPID_GetConfigInfo",
				 "malloc: SMP device name");
	      sprintf(new_config->device_init_name, "SMP_plug device");
	      new_config->num_served = smpi.num_local_nodes - 1;
	      new_config->granks_served = (int *)
		malloc(new_config->num_served * sizeof(int));
	      mxmpi_malloc_assert(new_config->granks_served,
				 "MPID_GetConfigInfo",
				 "malloc: SMP device map");
	      /* setup routes */
	      j = 0;
	      for(i=0; i<MPID_MyWorldSize; i++)
		{
		  if ((i!= MPID_MyWorldRank) 
		      && (smpi.local_nodes[i] != -1)) 
		    {
		      mxmpi_debug_assert(smpi.local_nodes[i] 
					!= smpi.my_local_id);
		      new_config->granks_served[j] = i;
		      j++;
		    }
		}
	      mxmpi_debug_assert(j == new_config->num_served);
	      new_config->next = NULL;
	    }
	}
      else
#endif
	{
	  smpi.num_local_nodes = 1;
	  for (i=0; i < MPID_MyWorldSize; i++)
	    {
	      if (i != MPID_MyWorldRank)
		{
		  smpi.local_nodes[i] = -1;
		}
	    }
	}
    }
  

  /* MX DEVICE */
  if (MPID_MyWorldSize > smpi.num_local_nodes)
    {
      if (new_config == NULL)
	{
	  new_config = (MPID_Config *)malloc(sizeof(MPID_Config));
	  mxmpi_malloc_assert(new_config,
			     "MPID_GetConfigInfo",
			     "malloc: MX device config");
	  return_config = new_config;
	}
      else
	{
	  new_config->next = (MPID_Config *)malloc(sizeof(MPID_Config));
	  mxmpi_malloc_assert(new_config->next,
			     "MPID_GetConfigInfo",
			     "malloc: MX device config");
	  new_config = new_config->next;
	}
      new_config->device_init = MPID_CH_InitMsgPass;
      new_config->device_init_name = (char *)malloc(255 * sizeof(char));
      mxmpi_malloc_assert(new_config->device_init_name,
			 "MPID_GetConfigInfo",
			 "malloc: MX device name");
      sprintf(new_config->device_init_name, "Myricom MX device");
      new_config->num_served = MPID_MyWorldSize - smpi.num_local_nodes;
      new_config->granks_served = (int *)malloc(new_config->num_served 
						* sizeof(int));
      mxmpi_malloc_assert(new_config->granks_served,
			 "MPID_GetConfigInfo",
			 "malloc: MX device map");
      /* setup the routes */
      j = 0;
      for(i=0; i<MPID_MyWorldSize; i++)
	if (smpi.local_nodes[i] == -1) {
	  new_config->granks_served[j] = i;
	  j++;
	}
      mxmpi_debug_assert(j == new_config->num_served);
      new_config->next = NULL;
    }
  
  return return_config;
}

