/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/

#include "mxmpi.h"

#if MX_DISABLE_REGISTRATION


void 
mxmpi_clear_interval(unsigned long start, unsigned int length)
{
  MXMPI_DEBUG_REG_CACHE_PRINT2("Clear_interval", "start",
			      start, "length", length);
}

void 
mxmpi_clear_all_intervals(void)
{
  MXMPI_DEBUG_REG_CACHE_PRINT0("Clear_all_intervals");
}

#else

#if MXMPI_ENABLE_REG_CACHE

typedef struct _entry
{
  unsigned long addr;
  struct _entry * prev;
  struct _entry * next;
  unsigned int refcount;
} regcache_entry;


static struct mx_hash * regcache_hash = NULL;
static struct mx_lookaside * regcache_lookaside = NULL;
static regcache_entry * regcache_head = NULL;
static regcache_entry * regcache_tail = NULL;


void 
mxmpi_regcache_init(void)
{
  mxmpi_debug_assert(MX_PAGE_LEN != 0);
  if (regcache_hash == NULL)
    {
      regcache_hash = mx_create_hash(mx_hash_compare_ptrs,
    				     mx_hash_hash_ptr, 0, 0,
				     4096, 0);
      regcache_lookaside = mx_create_lookaside(sizeof(regcache_entry),
					       4096);
    }

  mxmpi_malloc_assert(regcache_hash,
		     "mxmpi_regcache_init",
		     "mx_create_hash: regcache page hash");
  mxmpi_malloc_assert(regcache_lookaside,
		     "mxmpi_regcache_init",
		     "mx_create_lookaside: regcache entries list");
}


void 
mxmpi_regcache_deregister(void * addr, unsigned int pages)
{
  if (pages > 0)
    {
      mx_deregister_memory(mxmpi_mx_port, addr, MX_PAGE_LEN*pages);
      MXMPI_DEBUG_DMA_MEMORY_UNUSE(MX_PAGE_LEN*pages);
    }
}


void 
mxmpi_regcache_garbage_collector(unsigned int required)
{
  regcache_entry * entry_ptr, * next_entry;
  unsigned int count = 0;
  unsigned long batch_addr = 0;
  unsigned int batch_pages = 0;
  
  MXMPI_DEBUG_REG_CACHE_PRINT1("Garbage_collector start", "required", required);
  entry_ptr = regcache_head;
  while ((count < required) && (entry_ptr != NULL))
    {
      if (entry_ptr->refcount == 0)
	{
	  mx_hash_remove(regcache_hash, (void *)entry_ptr->addr);
	  if (batch_addr == 0)	  
	    {
	      batch_addr = entry_ptr->addr;
	      batch_pages++;
	    }
	  else
	    {
	      if (entry_ptr->addr == batch_addr+batch_pages*MX_PAGE_LEN)
		{
		  batch_pages++;
		}
	      else
		{
		  mxmpi_regcache_deregister((void *)batch_addr, batch_pages);
		  batch_addr = entry_ptr->addr;
		  batch_pages = 1;
		}
	    }
	  
	  count++;
	  next_entry = entry_ptr->next;
	  
	  if (regcache_head == entry_ptr)
	    regcache_head = next_entry;
	  else
	    entry_ptr->prev->next = next_entry;
	  
	  if (regcache_tail == entry_ptr)
	    regcache_tail = entry_ptr->prev;
	  else
	    entry_ptr->next->prev = entry_ptr->prev;
	  
	  mx_lookaside_free(entry_ptr);
          entry_ptr = next_entry;
        }
      else
        {
          entry_ptr = entry_ptr->next;
        }
    }
  
  if (batch_addr)
    {
      mxmpi_regcache_deregister((void *)batch_addr, batch_pages);
    }
  MXMPI_DEBUG_REG_CACHE_PRINT2("Garbage_collector stop", "required", 
			      required, "count", count);
}


unsigned int 
mxmpi_regcache_register(void * addr, unsigned int pages)
{
  unsigned int i;
  regcache_entry * entry_ptr;
  mx_status_t status;
  
  MXMPI_DEBUG_REG_CACHE_PRINT2("Regcache_register", "addr", 
			      (unsigned long)addr, "len", pages*MX_PAGE_LEN);
  
  if (mx_register_memory(mxmpi_mx_port, addr, MX_PAGE_LEN*pages) != MX_SUCCESS)
    {
      MXMPI_DEBUG_REG_CACHE_PRINT0("Regcache_register - using GC");
      mxmpi_regcache_garbage_collector(4096);
      if (mx_register_memory(mxmpi_mx_port, addr, MX_PAGE_LEN*pages)
	  != MX_SUCCESS)
	{
	  MXMPI_DEBUG_REG_CACHE_PRINT2("Register_memory failed", "start",
				      addr, "length", MX_PAGE_LEN*pages);
	  return 0;
	}
    }

  MXMPI_DEBUG_DMA_MEMORY_ACQUIRE(MX_PAGE_LEN*pages);
  
  for (i=0; i<pages; i++)
    {
      entry_ptr = (regcache_entry *)mx_lookaside_alloc(regcache_lookaside); 
      mxmpi_malloc_assert(entry_ptr,
			 "mxmpi_regcache_register",
			 "mx_lookaside_alloc: regcache entry");
			       
      if (regcache_head == NULL)
	{
	  regcache_head = entry_ptr;
	}
      else
        {     
          regcache_tail->next = entry_ptr;
        }
 
      entry_ptr->prev = regcache_tail;
      entry_ptr->next = NULL;
      regcache_tail = entry_ptr;
      entry_ptr->refcount = 1;
      MXMPI_DEBUG_DMA_MEMORY_USE(MX_PAGE_LEN);
      entry_ptr->addr = (unsigned long)addr + i*MX_PAGE_LEN;
      
      status = mx_hash_insert(regcache_hash,
                              (void *)(entry_ptr->addr),
			      (void *)(entry_ptr));
      if (status != MX_SUCCESS)
	{
	  fprintf(stderr, "[%d]: mx_hash_insert failure in "
		  "mxmpi_regcache_register: out of memory\n",
                  MPID_MyWorldRank);
	  mxmpi_abort (0);
	}
    }
  
  return 1;
}


unsigned long 
mxmpi_use_interval(unsigned long start, unsigned int length)
{
  unsigned long addr, end, batch_addr;
  unsigned int batch_pages;
  regcache_entry * entry_ptr;
 
  MXMPI_DEBUG_REG_CACHE_PRINT2("Use_interval", "start",
			      start, "len", length);
  if (length == 0)
    {
      return 0;
    }
  MXMPI_DEBUG_REGISTRATION_USE_SEGMENT(start, length);
  addr = start & ~(MX_PAGE_LEN-1);
  end = start + length;
  batch_addr = 0;
  batch_pages = 0;

  while (addr < end)
    {
      entry_ptr = (regcache_entry *)mx_hash_find(regcache_hash, (void *)addr);
      if (entry_ptr == NULL)
	{
	  if (batch_addr == 0)
	    {
	      batch_addr = addr;
	    }
	  batch_pages++;
	}
      else
	{	  
	  if (entry_ptr->refcount == 0)
	    {
	      MXMPI_DEBUG_DMA_MEMORY_USE(MX_PAGE_LEN);
	    }
	  
	  entry_ptr->refcount++;
          if (batch_addr != 0)
	    {
	      MXMPI_DEBUG_REG_CACHE_PRINT2("Use_interval batch", "batch_addr",
					  batch_addr, "batch_pages",
					  batch_pages);
	      if (mxmpi_regcache_register((void *)batch_addr, batch_pages) == 0)
		{
		  entry_ptr->refcount--;
		  
		  if (entry_ptr->refcount == 0)
		    {
		      MXMPI_DEBUG_DMA_MEMORY_UNUSE(MX_PAGE_LEN);
		      mem_locked_in_use -= MX_PAGE_LEN;
		    }
		  
		  if (batch_addr > start)
		    {
		      return (batch_addr-start);
		    }
		  else
		    {
		      return 0;
		    }
		}
	      
	      batch_addr = 0;
	      batch_pages = 0; 

	      /* move the entry to the end of the list (LRU policy) */
	      if (entry_ptr != regcache_tail)
		{
		  if (entry_ptr == regcache_head)
		    {
		      mxmpi_debug_assert(entry_ptr->next != NULL);
		      entry_ptr->next->prev = NULL;
		      regcache_head = entry_ptr->next;
		    }
		  else
		    {
		      mxmpi_debug_assert(entry_ptr->prev != NULL);
		      mxmpi_debug_assert(entry_ptr->next != NULL);
		      entry_ptr->prev->next = entry_ptr->next;
		      entry_ptr->next->prev = entry_ptr->prev;
		    }
		  
		  entry_ptr->next = NULL;
		  entry_ptr->prev = regcache_tail;
		  regcache_tail->next = entry_ptr;
                  regcache_tail = entry_ptr;
		}
	    }
	}
      addr += MX_PAGE_LEN;
    }
  
  if (batch_addr != 0)
    {
      if (mxmpi_regcache_register((void *)batch_addr, batch_pages) == 0)
	{
	  if (batch_addr > start)
	    {
	      return (batch_addr-start);
	    }
	  else
	    {
	      return 0;
	    }
	}
    }

  return length;
}


void 
mxmpi_unuse_interval(unsigned long start, unsigned int length)
{
  unsigned long addr, end;
  regcache_entry * entry_ptr;
  
  MXMPI_DEBUG_REG_CACHE_PRINT2("Unuse_interval", "start",
			      start, "length", length);
  if (length == 0)
    {
      return;
    }
  MXMPI_DEBUG_REGISTRATION_UNUSE_SEGMENT(start, length);

  addr = start & ~(MX_PAGE_LEN-1);
  end = start + length;
  
  while (addr < end)
    {
      entry_ptr = (regcache_entry *)mx_hash_find(regcache_hash, (void *)addr);
      
      mxmpi_debug_assert(entry_ptr != NULL);
      mxmpi_debug_assert(entry_ptr->refcount > 0);
      
      entry_ptr->refcount--;
      if (entry_ptr->refcount == 0)
	{
	  MXMPI_DEBUG_DMA_MEMORY_UNUSE(MX_PAGE_LEN); 
	}
      addr += MX_PAGE_LEN;
    }
}


void 
mxmpi_clear_interval(unsigned long start, unsigned int length)
{
  unsigned long addr, end, batch_addr;
  unsigned int batch_pages;
  regcache_entry * entry_ptr;
  
  MXMPI_DEBUG_REG_CACHE_PRINT2("Clear_interval", "start",
			      start, "length", length);
  MXMPI_DEBUG_REGISTRATION_CLEAR_SEGMENT(start, length);
				    
  if (regcache_hash != NULL)
    {
      addr = start & ~(MX_PAGE_LEN-1);
      end = start + length;
      batch_addr = 0;
      batch_pages = 0;
      
      while (addr < end)
	{
	  entry_ptr = (regcache_entry *)mx_hash_find(regcache_hash, 
						     (void *)addr);
	  if (entry_ptr != NULL)
	    {
	      mxmpi_debug_assert(entry_ptr->refcount == 0);
	      if (entry_ptr->refcount > 0)
		{
		  MXMPI_DEBUG_DMA_MEMORY_UNUSE(MX_PAGE_LEN);
		}
              mx_hash_remove(regcache_hash, (void *)addr);
	      
	      if (batch_addr == 0)
		batch_addr = addr;
	      batch_pages++;
	      
	      if (regcache_head == entry_ptr)
		regcache_head = entry_ptr->next;
	      else
		entry_ptr->prev->next = entry_ptr->next;
	  
	      if (regcache_tail == entry_ptr)
		regcache_tail = entry_ptr->prev;
	      else
		entry_ptr->next->prev = entry_ptr->prev;
	  
              mx_lookaside_free(entry_ptr);
	    }
	  else
	    {
	      if (batch_addr != 0)
		{
		  mxmpi_regcache_deregister((void *)batch_addr, batch_pages);
		  batch_addr = 0;
		  batch_pages = 0;
		}
	    }
	  addr += MX_PAGE_LEN;
	}

      if (batch_addr != 0)
	mxmpi_regcache_deregister((void *)batch_addr, batch_pages);
    }
}


void 
mxmpi_clear_all_intervals(void)
{
  struct mx_hash *old_regcache_hash;

  MXMPI_DEBUG_REG_CACHE_PRINT0("Clear_all_intervals");
  MXMPI_DEBUG_REGISTRATION_CLEAR_ALL_SEGMENTS();
  
  if (regcache_hash != NULL)
    {
      old_regcache_hash = regcache_hash;
      regcache_hash = NULL;
      mx_destroy_hash (old_regcache_hash);
      mx_destroy_lookaside (regcache_lookaside);
    }
}


#else /* NO_REG_CACHE */


unsigned long 
mxmpi_use_interval(unsigned long start, unsigned int length)
{
  MXMPI_DEBUG_REG_CACHE_PRINT2("Use_interval", "start",
			      start, "len", length);
  if (length == 0)
    {
      return 0;
    }
  MXMPI_DEBUG_REGISTRATION_USE_SEGMENT(start, length);
  
  if (mx_register_memory(mxmpi_mx_port, (void*)start, length) != MX_SUCCESS)
    {
      MXMPI_DEBUG_REG_CACHE_PRINT2("Use_interval no_regcache: register failed",
				  "start", start, "len", length);
      return 0;
    }
  
  MXMPI_DEBUG_DMA_MEMORY_USE(length);
  MXMPI_DEBUG_DMA_MEMORY_ACQUIRE(length);
  return length;
}

void 
mxmpi_unuse_interval(unsigned long start, unsigned int length)
{
  mx_status_t cc;

  MXMPI_DEBUG_REG_CACHE_PRINT2("Unuse_interval", "start",
			      start, "length", length);
  if (length == 0)
    {
      return;
    }
  MXMPI_DEBUG_REGISTRATION_UNUSE_SEGMENT(start, length);

  cc = mx_deregister_memory(mxmpi_mx_port, (void*)start, length);
  mxmpi_debug_assert(cc == MX_SUCCESS);

  MXMPI_DEBUG_DMA_MEMORY_UNUSE(length);
  MXMPI_DEBUG_DMA_MEMORY_RELEASE(length);
}


void 
mxmpi_regcache_init(void)
{
  ;
}

void 
mxmpi_clear_interval(unsigned long start, unsigned int length)
{
  MXMPI_DEBUG_REG_CACHE_PRINT2("Clear_interval", "start",
			      start, "length", length);
  MXMPI_DEBUG_REGISTRATION_CLEAR_SEGMENT(start, length);
}

void 
mxmpi_clear_all_intervals(void)
{
  MXMPI_DEBUG_REG_CACHE_PRINT0("Clear_all_intervals");
  MXMPI_DEBUG_REGISTRATION_CLEAR_ALL_SEGMENTS();
}

#endif

#endif
